import os
from mabledsocli.logger import Logger
from mabledsocli.config import Configs
from mabledsocli.providers import Providers
from mabledsocli.templates import TemplateProvider
from mabledsocli.stages import Stages
from mabledsocli.constants import *
from mabledsocli.exceptions import DSOException
from mabledsocli.contexts import Contexts
from mabledsocli.local_utils import *
from mabledsocli.settings import *


__default_spec = {
    'path': os.path.join(Configs.config_dir, 'templates'),
}


def get_default_spec():
    return __default_spec.copy()


class LocalTemplateProvider(TemplateProvider):


    def __init__(self):
        super().__init__('template/local/v1')


    @property
    def root_dir(self):
        return self.config.template_spec('path')


    def get_path_prefix(self):
        # return self.root_dir + os.sep
        return self.root_dir


    def list(self, config, uninherited=False, include_contents=False, filter=None):
        self.config = config
        Logger.debug(f"Listing local parameters: namespace={config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        templates = load_context_templates(config=config, path_prefix=self.get_path_prefix(), uninherited=uninherited, include_contents=include_contents, filter=filter)
        result = {'Templates': []}
        for key, details in templates.items():
            item = {'Key': key}
            item.update(details)
            result['Templates'].append(item)
        return result



    def add(self, config, key, contents, render_path=None):
        self.config = config
        if not Stages.is_default(config.stage) and not ALLOW_STAGE_TEMPLATES:
            raise DSOException(f"Templates may not be added to stage scopes, as the feature is currently disabled. It may be enabled by adding 'ALLOW_STAGE_TEMPLATES=yes' to the DSO global settings, or adding environment variable 'DSO_ALLOW_STAGE_TEMPLATES=yes'.")
        Logger.debug(f"Adding local template '{key}': namespace={config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        response = add_local_template(config=config, key=key, path_prefix=self.get_path_prefix(), contents=contents)
        result = {
                'Key': key,
                'Stage': Stages.shorten(config.stage),
                'Origin': response['Origin'], 
                'Path': response['Path'],
            }
        return result


    def get(self, config, key, revision=None):
        self.config = config
        if revision:
            raise DSOException(f"Template provider 'local/v1' does not support versioning.")
        Logger.debug(f"Getting template '{key}': namespace={config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        found = locate_template_in_context_hierachy(config=config, key=key, path_prefix=self.get_path_prefix(), include_contents=True)
        if not found:
            raise DSOException(f"Template '{key}' not found nor inherited in the given context: stage={config.short_stage}")
        result = {
                'Key': key, 
            }
        result.update(found[key])
        return result


    def delete(self, config, key):
        self.config = config
        Logger.debug(f"Locating template: namespace={config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        ### only parameters owned by the context can be deleted, hence uninherited=True
        found = locate_template_in_context_hierachy(config=config, key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: stage={Stages.shorten(stage)}")
        Logger.info(f"Deleting template: path={found[key]['Path']}")
        delete_local_template(path=found[key]['Path'])
        result = {
                'Key': key,
                'Stage': config.short_stage,
                'Scope': found[key]['Scope'], 
                'Origin': found[key]['Origin'], 
                'Path': found[key]['Path'],
            }
        return result


    def history(self, config, key, include_contents=False):
        raise DSOException(f"Template provider 'local/v1' does not support versioning.")



def register():
    Providers.register(LocalTemplateProvider())



from enum import Enum

class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented
    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented
    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    def __str__(self):
        return self.name.split('.')[-1]




import re
from .constants import *
from .exceptions import DSOException


REGEX_PATTERN = r"^([a-zA-Z][a-zA-Z0-9_-]+)/?([0-9])?$"

class StageService():

    # @staticmethod
    # def default_stage():
    #     return 'default/0'

    def raw_parse(self, stage):
        if not stage:
            import traceback
            traceback.print_stack() 
            raise DSOException(MESSAGES['Stage'].format(stage, REGEX_PATTERN))
        m = re.match(REGEX_PATTERN, stage)
        if m is None:
            raise DSOException(MESSAGES['InvalidStage'].format(stage, REGEX_PATTERN))
        stage = m.groups()[0]
        env = int(m.groups()[1]) if len(m.groups()) > 1 and m.groups()[1] else ''
        return stage, env

    def normalize(self, stage):
        stage, env = Stages.raw_parse(stage)
        # stage = stage or 'default'
        ### force dafault env if stage is default: default/env not allowed
        env = env if env and not stage == self.short_default_stage else 0
        return f"{stage}/{env}"

    def parse_normalized(self, stage):
        stage = Stages.normalize(stage)
        return Stages.raw_parse(stage)

    def get_default_env(self, stage):
        stage = Stages.parse_normalized(stage)[0]
        return f"{stage}/0"

    def parse_name(self, stage):
        return Stages.parse_normalized(stage)[0]

    def parse_env(self, stage):
        return Stages.parse_normalized(stage)[1]

    def shorten(self, stage):
        if not stage: return self.short_default_stage
        stage, env = Stages.parse_normalized(stage)
        if env == 0:
            return f"{stage}"
        else:
            return f"{stage}/{env}"

    @property
    def default_stage(self):
        return 'default/0'

    @property
    def short_default_stage(self):
        return 'default'

    def is_default(self, stage):
        return stage in [self.default_stage, self.short_default_stage]

    def is_default_env(self, stage):
        env = Stages.parse_normalized(stage)[1]
        return env == 0


Stages = StageService()


from .logger import Logger
from .constants import *
from .exceptions import *
from .network_utils import *

class NetworkService():
    def layout_subnet_plan(self, subnet_plan, filters=None, summary=True):
        Logger.info(f"Laying out subnet plan '{subnet_plan['name']}'...")
        return layout_subnet_plan(subnet_plan, filters, summary)


Networks = NetworkService()


from .cli import cli
from .version import __version__
from .exceptions import *
from .stages import Stages
from .contexts import Contexts
from .config import Configs
from .providers import ProviderBase, StoreProvider, ProviderManager, Providers
from .parameters import ParameterProvider, ParameterService, Parameters
from .secrets import SecretProvider, SecretService, Secrets
from .templates import TemplateProvider, TemplateService, Templates




#!/usr/bin/self.env python
import io
import os
import sys
from shutil import rmtree
from setuptools import find_packages, setup, Command
sys.path.append(os.path.dirname(__file__))
from version_utils import *
import requests
import json

# Package meta-data.
NAME = 'mabledsocli'
DESCRIPTION = 'Mable DSO CLI'
REQUIRES_PYTHON = '>=3.0'
URL = "https://github.com/bettercaring/mabledsocli"
EMAIL = 'ramtin.kazemi@mable.com.au'
AUTHOR = 'Ramtin Kazemi'
MAJOR_VERSION = 1
MINOR_VERSION = 0
PATCH_VERSION = -1  ### for automatic patch versioning set to -1

# ------------------------------------------------------------------------
# ------------------------------------------------------------------------
# ------------------------------------------------------------------------

here = os.path.abspath(os.path.dirname(__file__))

# ------------------------------------------------------------------------


### set_long_description
try:
    with io.open(os.path.join(here, 'PyPI-Description.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

prod_pypi_url = 'https://pypi.org/pypi/mabledsocli/json'
# dev_pypi_url='https://test.pypi.org/pypi/mabledsocli/json'
dev_pypi_url='https://pypi.org/pypi/mabledsocli/json'

### set_pypi_index
git_branch = os.popen('git symbolic-ref --short HEAD').read().strip()
if git_branch == 'master':
    env='prod'
    pypi_index='pypi'
elif git_branch == 'develop':
    env='dev'
    # pypi_index='testpypi'
    pypi_index='pypi'
else:
    raise Exception("Only 'develop' and 'master' git_branches can be deployed.")

### set version
try:
    dev_published_versions = list(reversed(sorted(requests.get(dev_pypi_url).json()['releases'].keys(), key=StrictVersionExt)))
    cur_dev_version = f"{MAJOR_VERSION}.{MINOR_VERSION}.{0 if PATCH_VERSION == -1 else PATCH_VERSION}.dev0"
    ### find the lastest matching prelease
    for v in dev_published_versions:
        _prerelease = parse_version(v)[3]
        if _prerelease == env:
            cur_dev_version = v
            break
except:
    cur_dev_version = f"{MAJOR_VERSION}.{MINOR_VERSION}.{0 if PATCH_VERSION == -1 else PATCH_VERSION}.dev0"

try:
    prod_published_versions = list(reversed(sorted(requests.get(prod_pypi_url).json()['releases'].keys(), key=StrictVersionExt)))
    cur_prod_version = f"{MAJOR_VERSION}.{MINOR_VERSION}.{0 if PATCH_VERSION == -1 else PATCH_VERSION}"
    ### find the lastest no prelease
    for v in prod_published_versions:
        _prerelease = parse_version(v)[3]
        if not _prerelease:
            cur_prod_version = v
            break
except:
    cur_prod_version = f"{MAJOR_VERSION}.{MINOR_VERSION}.{0 if PATCH_VERSION == -1 else PATCH_VERSION}"

if StrictVersionExt(cur_prod_version) > StrictVersionExt(cur_dev_version):
    cur_version = cur_prod_version
else:
    cur_version = cur_dev_version

cur_major, cur_minor, cur_patch, cur_prerelease, cur_prerelease_num = parse_version(cur_version)

new_major = MAJOR_VERSION
new_minor = MINOR_VERSION
new_prerelease = env if not env == 'prod' else None
new_prerelease_num = None

if new_major < cur_major:
    raise Exception(f"The new major version '{new_major}' cannot be smaller than the current one '{cur_major}'.")
elif new_major == cur_major:
    if new_minor < cur_minor:
        raise Exception(f"The new minor version '{new_minor}' cannot be smaller than the current one '{cur_minor}'.")
    elif new_minor == cur_minor:
        if new_prerelease:
            ### if lastest version is also prerelease, increase prerelease_num
            if cur_prerelease:
                new_patch = cur_patch
                new_prerelease_num = cur_prerelease_num + 1
            else:
                new_patch = cur_patch + 1 if PATCH_VERSION == -1 else PATCH_VERSION
                new_prerelease_num = 0
        ### no prerelease
        else:
            ### use smae pacth as prereleased if it is greater than the current in prod
            if StrictVersionExt(cur_dev_version) > StrictVersionExt(cur_prod_version):
                new_patch = cur_patch
            else:
                new_patch = cur_patch + 1 if PATCH_VERSION == -1 else PATCH_VERSION

            v1 = construct_version(cur_major, cur_minor, cur_patch, cur_prerelease, cur_prerelease_num)
            v2 = construct_version(new_major, new_minor, new_patch, new_prerelease, new_prerelease_num)
            if StrictVersionExt(v1) >= StrictVersionExt(v2):
                raise Exception(f"The new version '{v2}' must be greater than the current one '{v1}'.")
    ### minor version increased
    else:
        new_patch = 0 if PATCH_VERSION == -1 else PATCH_VERSION
### major version increased
else:
    new_patch = 0 if PATCH_VERSION == -1 else PATCH_VERSION

new_version = construct_version(new_major, new_minor, new_patch, new_prerelease, new_prerelease_num)

with open(os.path.join(here, 'src', NAME, 'package.json'), 'r') as f:
    package_info = json.load(f)
package_info['version'] = new_version
with open(os.path.join(here, 'src', NAME, 'package.json'), 'w') as f:
    json.dump(package_info, f, sort_keys=False, indent=2)

# ------------------------------------------------------------------------

class PublishCommand(Command):
    """Support setup.py publish."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass
        
    def run(self):
        
        self.status(f"Git Branch={git_branch}\nEnv={env}\nPyPI Index={pypi_index}\nCurrent Version={cur_version}\nNew Version={new_version}")
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
            rmtree(os.path.join(here, 'build'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        ### read from ~/.pypirc
        # TWINE_PASSWORD=$(<pypi-token.secret) python -m twine upload --username __token__ --skip-existing --repository testpypi dist/*
        code = os.system('{0} -m twine upload --skip-existing --repository {1} dist/*'.format(sys.executable, pypi_index))

        sys.exit(code)

        

# ------------------------------------------------------------------------
# ------------------------------------------------------------------------
# ------------------------------------------------------------------------

setup(
    name=NAME,
    version=new_version,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    project_urls={
        'Documentation': URL,
        'Source': URL
    },
    packages=find_packages('src', exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    package_dir = { '' : 'src/' }, 
    setup_requires=['wheel'],
    python_requires=REQUIRES_PYTHON,
    install_requires=open(os.path.join(here, 'requirements/prod.in'), 'r').readlines(),
    extras_require=None,
    package_data={'': ['package.json']},
    include_package_data=True,
    license='GPLV3',
    license_files = ['LICENSE.md'],
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    entry_points={
        # 'distutils.commands': ['publish = package.module:publish_command'],
        'console_scripts': [f'mabledso = {NAME}.cli:cli'],
        },
    cmdclass={'publish': PublishCommand},
)
